==> bodo-2022.7.1-cp310-cp310-manylinux2014_x86_64.manylinux_2_17_x86_64.whl: bodo/utils/indexing.py <==
"""
Collection of utility functions for indexing implementation (getitem/setitem)
"""
import operator
import numba
import numpy as np
from numba.core import types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import intrinsic, overload, register_jitable
import bodo
from bodo.utils.typing import BodoError


@register_jitable
def get_new_null_mask_bool_index(old_mask, ind, n):
    tsdt__nlmcb = n + 7 >> 3
    tytfw__jpypi = np.empty(tsdt__nlmcb, np.uint8)
    okb__jbyc = 0
    for hfbk__spi in range(len(ind)):
        if ind[hfbk__spi]:
            uvbqt__ddog = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
                hfbk__spi)
            bodo.libs.int_arr_ext.set_bit_to_arr(tytfw__jpypi, okb__jbyc,
                uvbqt__ddog)
            okb__jbyc += 1
    return tytfw__jpypi


@register_jitable
def array_getitem_bool_index(A, ind):
    ind = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    vrgw__guujq = A._data[ind]
    n = len(vrgw__guujq)
    tytfw__jpypi = get_new_null_mask_bool_index(old_mask, ind, n)
    return vrgw__guujq, tytfw__jpypi


@register_jitable
def get_new_null_mask_int_index(old_mask, ind, n):
    tsdt__nlmcb = n + 7 >> 3
    tytfw__jpypi = np.empty(tsdt__nlmcb, np.uint8)
    okb__jbyc = 0
    for hfbk__spi in range(len(ind)):
        uvbqt__ddog = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            ind[hfbk__spi])
        bodo.libs.int_arr_ext.set_bit_to_arr(tytfw__jpypi, okb__jbyc,
            uvbqt__ddog)
        okb__jbyc += 1
    return tytfw__jpypi


@register_jitable
def array_getitem_int_index(A, ind):
    ddf__doz = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    vrgw__guujq = A._data[ddf__doz]
    n = len(vrgw__guujq)
    tytfw__jpypi = get_new_null_mask_int_index(old_mask, ddf__doz, n)
    return vrgw__guujq, tytfw__jpypi


@register_jitable
def get_new_null_mask_slice_index(old_mask, ind, n):
    jca__fcbz = numba.cpython.unicode._normalize_slice(ind, n)
    bdn__rgs = numba.cpython.unicode._slice_span(jca__fcbz)
    tsdt__nlmcb = bdn__rgs + 7 >> 3
    tytfw__jpypi = np.empty(tsdt__nlmcb, np.uint8)
    okb__jbyc = 0
    for hfbk__spi in range(jca__fcbz.start, jca__fcbz.stop, jca__fcbz.step):
        uvbqt__ddog = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            hfbk__spi)
        bodo.libs.int_arr_ext.set_bit_to_arr(tytfw__jpypi, okb__jbyc,
            uvbqt__ddog)
        okb__jbyc += 1
    return tytfw__jpypi


@register_jitable
def array_getitem_slice_index(A, ind):
    n = len(A._data)
    old_mask = A._null_bitmap
    vrgw__guujq = np.ascontiguousarray(A._data[ind])
    tytfw__jpypi = get_new_null_mask_slice_index(old_mask, ind, n)
    return vrgw__guujq, tytfw__jpypi


def array_setitem_int_index(A, idx, val):
    return


@overload(array_setitem_int_index, no_unliteral=True)
def array_setitem_int_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(val._data)
            for hfbk__spi in range(n):
                A._data[idx[hfbk__spi]] = val._data[hfbk__spi]
                uvbqt__ddog = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val.
                    _null_bitmap, hfbk__spi)
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, idx[
                    hfbk__spi], uvbqt__ddog)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            for hfbk__spi in idx:
                A._data[hfbk__spi] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    hfbk__spi, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def array_setitem_bool_index(A, idx, val):
    A[idx] = val


@overload(array_setitem_bool_index, no_unliteral=True)
def array_setitem_bool_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(idx)
            mncd__vlj = 0
            for hfbk__spi in range(n):
                if not bodo.libs.array_kernels.isna(idx, hfbk__spi) and idx[
                    hfbk__spi]:
                    A._data[hfbk__spi] = val._data[mncd__vlj]
                    uvbqt__ddog = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                        ._null_bitmap, mncd__vlj)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        hfbk__spi, uvbqt__ddog)
                    mncd__vlj += 1
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            n = len(idx)
            mncd__vlj = 0
            for hfbk__spi in range(n):
                if not bodo.libs.array_kernels.isna(idx, hfbk__spi) and idx[
                    hfbk__spi]:
                    A._data[hfbk__spi] = val
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        hfbk__spi, 1)
                    mncd__vlj += 1
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


@register_jitable
def setitem_slice_index_null_bits(dst_bitmap, src_bitmap, idx, n):
    jca__fcbz = numba.cpython.unicode._normalize_slice(idx, n)
    mncd__vlj = 0
    for hfbk__spi in range(jca__fcbz.start, jca__fcbz.stop, jca__fcbz.step):
        uvbqt__ddog = bodo.libs.int_arr_ext.get_bit_bitmap_arr(src_bitmap,
            mncd__vlj)
        bodo.libs.int_arr_ext.set_bit_to_arr(dst_bitmap, hfbk__spi, uvbqt__ddog
            )
        mncd__vlj += 1


def array_setitem_slice_index(A, idx, val):
    return


@overload(array_setitem_slice_index, no_unliteral=True)
def array_setitem_slice_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(A._data)
            A._data[idx] = val._data
            src_bitmap = val._null_bitmap.copy()
            setitem_slice_index_null_bits(A._null_bitmap, src_bitmap, idx, n)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            jca__fcbz = numba.cpython.unicode._normalize_slice(idx, len(A))
            for hfbk__spi in range(jca__fcbz.start, jca__fcbz.stop,
                jca__fcbz.step):
                A._data[hfbk__spi] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    hfbk__spi, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def untuple_if_one_tuple(v):
    return v


@overload(untuple_if_one_tuple)
def untuple_if_one_tuple_overload(v):
    if isinstance(v, types.BaseTuple) and len(v.types) == 1:
        return lambda v: v[0]
    return lambda v: v


def init_nested_counts(arr_typ):
    return 0,


@overload(init_nested_counts)
def overload_init_nested_counts(arr_typ):
    arr_typ = arr_typ.instance_type
    if isinstance(arr_typ, bodo.libs.array_item_arr_ext.ArrayItemArrayType
        ) or arr_typ == bodo.string_array_type:
        data_arr_typ = arr_typ.dtype
        return lambda arr_typ: (0,) + init_nested_counts(data_arr_typ)
    if bodo.utils.utils.is_array_typ(arr_typ, False
        ) or arr_typ == bodo.string_type:
        return lambda arr_typ: (0,)
    return lambda arr_typ: ()


def add_nested_counts(nested_counts, arr_item):
    return 0,


@overload(add_nested_counts)
def overload_add_nested_counts(nested_counts, arr_item):
    from bodo.libs.str_arr_ext import get_utf8_size
    arr_item = arr_item.type if isinstance(arr_item, types.Optional
        ) else arr_item
    if isinstance(arr_item, bodo.libs.array_item_arr_ext.ArrayItemArrayType):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),) + add_nested_counts(nested_counts[1:], bodo.libs.
            array_item_arr_ext.get_data(arr_item))
    if isinstance(arr_item, types.List):
        return lambda nested_counts, arr_item: add_nested_counts(nested_counts,
            bodo.utils.conversion.coerce_to_array(arr_item))
    if arr_item == bodo.string_array_type:
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item), nested_counts[1] + np.int64(bodo.libs.str_arr_ext.
            num_total_chars(arr_item)))
    if bodo.utils.utils.is_array_typ(arr_item, False):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),)
    if arr_item == bodo.string_type:
        return lambda nested_counts, arr_item: (nested_counts[0] +
            get_utf8_size(arr_item),)
    return lambda nested_counts, arr_item: ()


@overload(operator.setitem)
def none_optional_setitem_overload(A, idx, val):
    if not bodo.utils.utils.is_array_typ(A, False):
        return
    elif val == types.none:
        if isinstance(idx, types.Integer):
            return lambda A, idx, val: bodo.libs.array_kernels.setna(A, idx)
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_none_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for hfbk__spi in idx:
                    bodo.libs.array_kernels.setna(A, hfbk__spi)
            return setitem_none_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    n = len(A)
                    idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                    alw__jwim = bodo.libs.str_arr_ext.pre_alloc_string_array(n,
                        -1)
                    for hfbk__spi in numba.parfors.parfor.internal_prange(n):
                        if idx[hfbk__spi] or bodo.libs.array_kernels.isna(A,
                            hfbk__spi):
                            alw__jwim[hfbk__spi] = ''
                            bodo.libs.str_arr_ext.str_arr_set_na(alw__jwim,
                                hfbk__spi)
                        else:
                            alw__jwim[hfbk__spi] = A[hfbk__spi]
                    bodo.libs.str_arr_ext.move_str_binary_arr_payload(A,
                        alw__jwim)
                return string_arr_impl

            def setitem_none_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for hfbk__spi in range(n):
                    if not bodo.libs.array_kernels.isna(idx, hfbk__spi
                        ) and idx[hfbk__spi]:
                        bodo.libs.array_kernels.setna(A, hfbk__spi)
            return setitem_none_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_none_slice(A, idx, val):
                n = len(A)
                jca__fcbz = numba.cpython.unicode._normalize_slice(idx, n)
                for hfbk__spi in range(jca__fcbz.start, jca__fcbz.stop,
                    jca__fcbz.step):
                    bodo.libs.array_kernels.setna(A, hfbk__spi)
            return setitem_none_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and None value not supported.'
            )
    elif isinstance(val, types.optional):
        if isinstance(idx, types.Integer):

            def impl_optional(A, idx, val):
                if val is None:
                    bodo.libs.array_kernels.setna(A, idx)
                else:
                    A[idx] = bodo.utils.indexing.unoptional(val)
            return impl_optional
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_optional_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for hfbk__spi in idx:
                    if val is None:
                        bodo.libs.array_kernels.setna(A, hfbk__spi)
                        continue
                    A[hfbk__spi] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    if val is None:
                        A[idx] = None
                    else:
                        A[idx] = bodo.utils.indexing.unoptional(val)
                return string_arr_impl

            def setitem_optional_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for hfbk__spi in range(n):
                    if not bodo.libs.array_kernels.isna(idx, hfbk__spi
                        ) and idx[hfbk__spi]:
                        if val is None:
                            bodo.libs.array_kernels.setna(A, hfbk__spi)
                            continue
                        A[hfbk__spi] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_optional_slice(A, idx, val):
                n = len(A)
                jca__fcbz = numba.cpython.unicode._normalize_slice(idx, n)
                for hfbk__spi in range(jca__fcbz.start, jca__fcbz.stop,
                    jca__fcbz.step):
                    if val is None:
                        bodo.libs.array_kernels.setna(A, hfbk__spi)
                        continue
                    A[hfbk__spi] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and optional value not supported.'
            )


@intrinsic
def unoptional(typingctx, val_t=None):
    if not isinstance(val_t, types.Optional):
        return val_t(val_t), lambda c, b, s, args: impl_ret_borrowed(c, b,
            val_t, args[0])

    def codegen(context, builder, signature, args):
        fmt__vmg = context.make_helper(builder, val_t, args[0])
        kfgi__asl = fmt__vmg.data
        context.nrt.incref(builder, val_t.type, kfgi__asl)
        return kfgi__asl
    return val_t.type(val_t), codegen

==> bodo-2022.7.3-cp310-cp310-manylinux2014_x86_64.manylinux_2_17_x86_64.whl: bodo/utils/indexing.py <==
"""
Collection of utility functions for indexing implementation (getitem/setitem)
"""
import operator
import numba
import numpy as np
from numba.core import types
from numba.core.imputils import impl_ret_borrowed
from numba.extending import intrinsic, overload, register_jitable
import bodo
from bodo.utils.typing import BodoError


@register_jitable
def get_new_null_mask_bool_index(old_mask, ind, n):
    jze__ihamp = n + 7 >> 3
    kswp__alydk = np.empty(jze__ihamp, np.uint8)
    bzx__nmx = 0
    for ofsc__vrtn in range(len(ind)):
        if ind[ofsc__vrtn]:
            wkvr__wfzno = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
                ofsc__vrtn)
            bodo.libs.int_arr_ext.set_bit_to_arr(kswp__alydk, bzx__nmx,
                wkvr__wfzno)
            bzx__nmx += 1
    return kswp__alydk


@register_jitable
def array_getitem_bool_index(A, ind):
    ind = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    ypqs__fipa = A._data[ind]
    n = len(ypqs__fipa)
    kswp__alydk = get_new_null_mask_bool_index(old_mask, ind, n)
    return ypqs__fipa, kswp__alydk


@register_jitable
def get_new_null_mask_int_index(old_mask, ind, n):
    jze__ihamp = n + 7 >> 3
    kswp__alydk = np.empty(jze__ihamp, np.uint8)
    bzx__nmx = 0
    for ofsc__vrtn in range(len(ind)):
        wkvr__wfzno = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            ind[ofsc__vrtn])
        bodo.libs.int_arr_ext.set_bit_to_arr(kswp__alydk, bzx__nmx, wkvr__wfzno
            )
        bzx__nmx += 1
    return kswp__alydk


@register_jitable
def array_getitem_int_index(A, ind):
    mcukk__tbx = bodo.utils.conversion.coerce_to_ndarray(ind)
    old_mask = A._null_bitmap
    ypqs__fipa = A._data[mcukk__tbx]
    n = len(ypqs__fipa)
    kswp__alydk = get_new_null_mask_int_index(old_mask, mcukk__tbx, n)
    return ypqs__fipa, kswp__alydk


@register_jitable
def get_new_null_mask_slice_index(old_mask, ind, n):
    zlkw__bddp = numba.cpython.unicode._normalize_slice(ind, n)
    bia__awgi = numba.cpython.unicode._slice_span(zlkw__bddp)
    jze__ihamp = bia__awgi + 7 >> 3
    kswp__alydk = np.empty(jze__ihamp, np.uint8)
    bzx__nmx = 0
    for ofsc__vrtn in range(zlkw__bddp.start, zlkw__bddp.stop, zlkw__bddp.step
        ):
        wkvr__wfzno = bodo.libs.int_arr_ext.get_bit_bitmap_arr(old_mask,
            ofsc__vrtn)
        bodo.libs.int_arr_ext.set_bit_to_arr(kswp__alydk, bzx__nmx, wkvr__wfzno
            )
        bzx__nmx += 1
    return kswp__alydk


@register_jitable
def array_getitem_slice_index(A, ind):
    n = len(A._data)
    old_mask = A._null_bitmap
    ypqs__fipa = np.ascontiguousarray(A._data[ind])
    kswp__alydk = get_new_null_mask_slice_index(old_mask, ind, n)
    return ypqs__fipa, kswp__alydk


def array_setitem_int_index(A, idx, val):
    return


@overload(array_setitem_int_index, no_unliteral=True)
def array_setitem_int_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(val._data)
            for ofsc__vrtn in range(n):
                A._data[idx[ofsc__vrtn]] = val._data[ofsc__vrtn]
                wkvr__wfzno = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val.
                    _null_bitmap, ofsc__vrtn)
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap, idx[
                    ofsc__vrtn], wkvr__wfzno)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            for ofsc__vrtn in idx:
                A._data[ofsc__vrtn] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    ofsc__vrtn, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def array_setitem_bool_index(A, idx, val):
    A[idx] = val


@overload(array_setitem_bool_index, no_unliteral=True)
def array_setitem_bool_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(idx)
            tefuu__ulp = 0
            for ofsc__vrtn in range(n):
                if not bodo.libs.array_kernels.isna(idx, ofsc__vrtn) and idx[
                    ofsc__vrtn]:
                    A._data[ofsc__vrtn] = val._data[tefuu__ulp]
                    wkvr__wfzno = bodo.libs.int_arr_ext.get_bit_bitmap_arr(val
                        ._null_bitmap, tefuu__ulp)
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        ofsc__vrtn, wkvr__wfzno)
                    tefuu__ulp += 1
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            n = len(idx)
            tefuu__ulp = 0
            for ofsc__vrtn in range(n):
                if not bodo.libs.array_kernels.isna(idx, ofsc__vrtn) and idx[
                    ofsc__vrtn]:
                    A._data[ofsc__vrtn] = val
                    bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                        ofsc__vrtn, 1)
                    tefuu__ulp += 1
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


@register_jitable
def setitem_slice_index_null_bits(dst_bitmap, src_bitmap, idx, n):
    zlkw__bddp = numba.cpython.unicode._normalize_slice(idx, n)
    tefuu__ulp = 0
    for ofsc__vrtn in range(zlkw__bddp.start, zlkw__bddp.stop, zlkw__bddp.step
        ):
        wkvr__wfzno = bodo.libs.int_arr_ext.get_bit_bitmap_arr(src_bitmap,
            tefuu__ulp)
        bodo.libs.int_arr_ext.set_bit_to_arr(dst_bitmap, ofsc__vrtn,
            wkvr__wfzno)
        tefuu__ulp += 1


def array_setitem_slice_index(A, idx, val):
    return


@overload(array_setitem_slice_index, no_unliteral=True)
def array_setitem_slice_index_overload(A, idx, val):
    if bodo.utils.utils.is_array_typ(val
        ) or bodo.utils.typing.is_iterable_type(val):

        def impl_arr(A, idx, val):
            val = bodo.utils.conversion.coerce_to_array(val,
                use_nullable_array=True)
            n = len(A._data)
            A._data[idx] = val._data
            src_bitmap = val._null_bitmap.copy()
            setitem_slice_index_null_bits(A._null_bitmap, src_bitmap, idx, n)
        return impl_arr
    if bodo.utils.typing.is_scalar_type(val):

        def impl_scalar(A, idx, val):
            zlkw__bddp = numba.cpython.unicode._normalize_slice(idx, len(A))
            for ofsc__vrtn in range(zlkw__bddp.start, zlkw__bddp.stop,
                zlkw__bddp.step):
                A._data[ofsc__vrtn] = val
                bodo.libs.int_arr_ext.set_bit_to_arr(A._null_bitmap,
                    ofsc__vrtn, 1)
        return impl_scalar
    raise BodoError(f'setitem not supported for {A} with value {val}')


def untuple_if_one_tuple(v):
    return v


@overload(untuple_if_one_tuple)
def untuple_if_one_tuple_overload(v):
    if isinstance(v, types.BaseTuple) and len(v.types) == 1:
        return lambda v: v[0]
    return lambda v: v


def init_nested_counts(arr_typ):
    return 0,


@overload(init_nested_counts)
def overload_init_nested_counts(arr_typ):
    arr_typ = arr_typ.instance_type
    if isinstance(arr_typ, bodo.libs.array_item_arr_ext.ArrayItemArrayType
        ) or arr_typ == bodo.string_array_type:
        data_arr_typ = arr_typ.dtype
        return lambda arr_typ: (0,) + init_nested_counts(data_arr_typ)
    if bodo.utils.utils.is_array_typ(arr_typ, False
        ) or arr_typ == bodo.string_type:
        return lambda arr_typ: (0,)
    return lambda arr_typ: ()


def add_nested_counts(nested_counts, arr_item):
    return 0,


@overload(add_nested_counts)
def overload_add_nested_counts(nested_counts, arr_item):
    from bodo.libs.str_arr_ext import get_utf8_size
    arr_item = arr_item.type if isinstance(arr_item, types.Optional
        ) else arr_item
    if isinstance(arr_item, bodo.libs.array_item_arr_ext.ArrayItemArrayType):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),) + add_nested_counts(nested_counts[1:], bodo.libs.
            array_item_arr_ext.get_data(arr_item))
    if isinstance(arr_item, types.List):
        return lambda nested_counts, arr_item: add_nested_counts(nested_counts,
            bodo.utils.conversion.coerce_to_array(arr_item))
    if arr_item == bodo.string_array_type:
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item), nested_counts[1] + np.int64(bodo.libs.str_arr_ext.
            num_total_chars(arr_item)))
    if bodo.utils.utils.is_array_typ(arr_item, False):
        return lambda nested_counts, arr_item: (nested_counts[0] + len(
            arr_item),)
    if arr_item == bodo.string_type:
        return lambda nested_counts, arr_item: (nested_counts[0] +
            get_utf8_size(arr_item),)
    return lambda nested_counts, arr_item: ()


@overload(operator.setitem)
def none_optional_setitem_overload(A, idx, val):
    if not bodo.utils.utils.is_array_typ(A, False):
        return
    elif val == types.none:
        if isinstance(idx, types.Integer):
            return lambda A, idx, val: bodo.libs.array_kernels.setna(A, idx)
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_none_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for ofsc__vrtn in idx:
                    bodo.libs.array_kernels.setna(A, ofsc__vrtn)
            return setitem_none_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    n = len(A)
                    idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                    liu__brab = bodo.libs.str_arr_ext.pre_alloc_string_array(n,
                        -1)
                    for ofsc__vrtn in numba.parfors.parfor.internal_prange(n):
                        if idx[ofsc__vrtn] or bodo.libs.array_kernels.isna(A,
                            ofsc__vrtn):
                            liu__brab[ofsc__vrtn] = ''
                            bodo.libs.str_arr_ext.str_arr_set_na(liu__brab,
                                ofsc__vrtn)
                        else:
                            liu__brab[ofsc__vrtn] = A[ofsc__vrtn]
                    bodo.libs.str_arr_ext.move_str_binary_arr_payload(A,
                        liu__brab)
                return string_arr_impl

            def setitem_none_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for ofsc__vrtn in range(n):
                    if not bodo.libs.array_kernels.isna(idx, ofsc__vrtn
                        ) and idx[ofsc__vrtn]:
                        bodo.libs.array_kernels.setna(A, ofsc__vrtn)
            return setitem_none_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_none_slice(A, idx, val):
                n = len(A)
                zlkw__bddp = numba.cpython.unicode._normalize_slice(idx, n)
                for ofsc__vrtn in range(zlkw__bddp.start, zlkw__bddp.stop,
                    zlkw__bddp.step):
                    bodo.libs.array_kernels.setna(A, ofsc__vrtn)
            return setitem_none_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and None value not supported.'
            )
    elif isinstance(val, types.optional):
        if isinstance(idx, types.Integer):

            def impl_optional(A, idx, val):
                if val is None:
                    bodo.libs.array_kernels.setna(A, idx)
                else:
                    A[idx] = bodo.utils.indexing.unoptional(val)
            return impl_optional
        elif bodo.utils.typing.is_list_like_index_type(idx) and isinstance(idx
            .dtype, types.Integer):

            def setitem_optional_int_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                for ofsc__vrtn in idx:
                    if val is None:
                        bodo.libs.array_kernels.setna(A, ofsc__vrtn)
                        continue
                    A[ofsc__vrtn] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_int_arr
        elif bodo.utils.typing.is_list_like_index_type(idx
            ) and idx.dtype == types.bool_:
            if A == bodo.string_array_type:

                def string_arr_impl(A, idx, val):
                    if val is None:
                        A[idx] = None
                    else:
                        A[idx] = bodo.utils.indexing.unoptional(val)
                return string_arr_impl

            def setitem_optional_bool_arr(A, idx, val):
                idx = bodo.utils.conversion.coerce_to_ndarray(idx)
                n = len(idx)
                for ofsc__vrtn in range(n):
                    if not bodo.libs.array_kernels.isna(idx, ofsc__vrtn
                        ) and idx[ofsc__vrtn]:
                        if val is None:
                            bodo.libs.array_kernels.setna(A, ofsc__vrtn)
                            continue
                        A[ofsc__vrtn] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_bool_arr
        elif isinstance(idx, types.SliceType):

            def setitem_optional_slice(A, idx, val):
                n = len(A)
                zlkw__bddp = numba.cpython.unicode._normalize_slice(idx, n)
                for ofsc__vrtn in range(zlkw__bddp.start, zlkw__bddp.stop,
                    zlkw__bddp.step):
                    if val is None:
                        bodo.libs.array_kernels.setna(A, ofsc__vrtn)
                        continue
                    A[ofsc__vrtn] = bodo.utils.indexing.unoptional(val)
            return setitem_optional_slice
        raise BodoError(
            f'setitem for {A} with indexing type {idx} and optional value not supported.'
            )


@intrinsic
def unoptional(typingctx, val_t=None):
    if not isinstance(val_t, types.Optional):
        return val_t(val_t), lambda c, b, s, args: impl_ret_borrowed(c, b,
            val_t, args[0])

    def codegen(context, builder, signature, args):
        fwy__bnlbz = context.make_helper(builder, val_t, args[0])
        cspe__trtu = fwy__bnlbz.data
        context.nrt.incref(builder, val_t.type, cspe__trtu)
        return cspe__trtu
    return val_t.type(val_t), codegen
